// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.8 Include File. (povr version ONLY)
// File: setidtypes.inc
// Last updated: 23 September 2021

// Description: Creates global.dictIdTypes with some common type values for
// the running version of a povr branch of POV-Ray.

// The povr branch adds a new keyword in id_type() which can determine numeric
// values (internal enums) for #local, #declare and passed macro parameter
// generated IDs. These enum/integer values come mostly from an internal enum
// table which is not stable as keywords change, but which can be used to do
// some identifier type checking.

// Also defines macros:
//
// Mcr_id_type()  - Used to create internal ID from raw SDL.
// McrPrintGlobalDictIdTypes(true) - Dump of ID information and some help text.
//

#macro Mcr_id_type(_id) id_type(_id) #end

#declare dictIdTypes = dictionary {
    .boo: Mcr_id_type(true),
    .int: Mcr_id_type(1),
    .flt: Mcr_id_type(1.0),
    .v2d: Mcr_id_type(<1,2>),
    .v3d: Mcr_id_type(<1,2,3>),
    .v4d: Mcr_id_type(<1,2,3,4>),
    .v5d: Mcr_id_type(<1,2,3,4,5>),
    .str: Mcr_id_type("Dear dictidtypes"),
    .obj: Mcr_id_type(sphere{0,1}),
    .csg: Mcr_id_type(union{object{sphere{0,1}} object{sphere{1,1}}}),
    .l_s: Mcr_id_type(light_source{<3,3,3> rgb 1}),
    .l_g: Mcr_id_type(light_group{light_source{<3,3,3> rgb 1} light_source{<7,7,7> rgb 1}}),
    .l_u: Mcr_id_type(union{light_source{<3,3,3> rgb 1} light_source{<7,7,7> rgb 1}}),
    .ary: Mcr_id_type(array mixed),
    .dic: Mcr_id_type(dictionary),
    .fnf: Mcr_id_type(function { 1 }),
    .fnv: Mcr_id_type(function { pigment { rgb 1 } }),
    .fin: Mcr_id_type(finish { emission 1 }),
    .nrm: Mcr_id_type(normal { wrinkles }),
    .pig: Mcr_id_type(pigment { color rgb 1 }),
    .txr: Mcr_id_type(texture {pigment { color rgb 1}}),
    .itr: Mcr_id_type(interior {ior 1.3}),
    .mat: Mcr_id_type(material {texture {pigment { color rgb 1}} interior {ior 1.3}}),
    .den: Mcr_id_type(density { function { 1 } }),
    .dem: Mcr_id_type(density_map {[0 rgb 0][1 rgb 1]}),
    .med: Mcr_id_type(media { density { function { 1 } } }),
    .fog: Mcr_id_type(fog { distance 100.0 }),
    .clm: Mcr_id_type(color_map {[0 rgb 0][1 rgb 1]}),
    .pgm: Mcr_id_type(pigment_map {[0 rgb 0][1 rgb 1]}),
    .trm: Mcr_id_type(transform { rotate x*14 }),
    .cam: Mcr_id_type(camera {location 3}),
    .mcr: id_type(global.Mcr_id_type),
    .spl: Mcr_id_type(spline { 1, <1,1>, 2, <2,2>, 3, <3,3>}),
    .sky: Mcr_id_type(sky_sphere { pigment { color rgb 1 } }),

}

#macro McrPrintGlobalDictIdTypes(optional _tf)
    #ifndef (local._tf)
        #local TF = false;
    #else
        #local TF = _tf;
    #end

    #if (TF)
        #debug "\n-------\n"
        #debug "Notes. In POV-Ray booleans and integers are floats with specific values.\n"
        #debug "Booleans have 0.0 or 1.0 values for false and true.\n"
        #debug "Integers are floats with no significant decimal digits. 1, 1.0, 1.00\n"
        #debug "-\n"
        #debug "Internally POV-Ray's object category is very broad. It includes all\n"
        #debug "surfaces, objects having a defined inside, all csg, light sources,\n"
        #debug "light unions and light groups.\n"
        #debug "-\n"
        #debug "The id_type() function breaks out csg, light sources, light unions and\n"
        #debug "light groups with individual negative values so additional type checking\n"
        #debug "can be done. To test specifically for POV-Ray's more general object\n"
        #debug "types and compound test must be used as in:\n\n"
        #debug "    #if ((id_type(Thing00)=global.dictIdTypes.obj) |\n"
        #debug "         (id_type(Thing00)<0.0))\n\n"
        #debug "-\n"
        #debug "There is no difference in type id between 'array' and 'array mixed'.\n"
        #debug "-\n"
        #debug "WARNING! While type integer values are shown below. Do NOT use them\n"
        #debug "         directly. These values often change release to release.\n"
        #debug "---\n"
        #debug concat("global.dictIdTypes.boo --> ",str(global.dictIdTypes.int,4,0)," booleans\n")
        #debug concat("global.dictIdTypes.int --> ",str(global.dictIdTypes.int,4,0)," integers\n")
        #debug concat("global.dictIdTypes.flt --> ",str(global.dictIdTypes.flt,4,0)," floats\n")
        #debug concat("global.dictIdTypes.v2d --> ",str(global.dictIdTypes.v2d,4,0)," two dimension vector\n")
        #debug concat("global.dictIdTypes.v3d --> ",str(global.dictIdTypes.v3d,4,0)," three dimension vector\n")
        #debug concat("global.dictIdTypes.v4d --> ",str(global.dictIdTypes.v4d,4,0)," four dimension vector\n")
        #debug concat("global.dictIdTypes.v5d --> ",str(global.dictIdTypes.v5d,4,0)," five dimension vector\n")
        #debug concat("global.dictIdTypes.str --> ",str(global.dictIdTypes.str,4,0)," string\n")
        #debug concat("global.dictIdTypes.obj --> ",str(global.dictIdTypes.obj,4,0)," simple surface, object with inside\n")
        #debug concat("global.dictIdTypes.csg --> ",str(global.dictIdTypes.csg,4,0)," constructive solid geometry. Union, etc\n")
        #debug concat("global.dictIdTypes.l_s --> ",str(global.dictIdTypes.l_s,4,0)," light_source\n")
        #debug concat("global.dictIdTypes.l_g --> ",str(global.dictIdTypes.l_g,4,0)," light_group\n")
        #debug concat("global.dictIdTypes.l_u --> ",str(global.dictIdTypes.l_u,4,0)," union of light_source(s)\n")
        #debug concat("global.dictIdTypes.ary --> ",str(global.dictIdTypes.ary,4,0)," array\n")
        #debug concat("global.dictIdTypes.dic --> ",str(global.dictIdTypes.dic,4,0)," dictionary\n")
        #debug concat("global.dictIdTypes.fnf --> ",str(global.dictIdTypes.fnf,4,0)," function returning float\n")
        #debug concat("global.dictIdTypes.fnv --> ",str(global.dictIdTypes.fnv,4,0)," function returning vector\n")
        #debug concat("global.dictIdTypes.fin --> ",str(global.dictIdTypes.fin,4,0)," finish\n")
        #debug concat("global.dictIdTypes.nrm --> ",str(global.dictIdTypes.nrm,4,0)," normal\n")
        #debug concat("global.dictIdTypes.pig --> ",str(global.dictIdTypes.pig,4,0)," pigment\n")
        #debug concat("global.dictIdTypes.txr --> ",str(global.dictIdTypes.txr,4,0)," texture\n")
        #debug concat("global.dictIdTypes.itr --> ",str(global.dictIdTypes.itr,4,0)," interior\n")
        #debug concat("global.dictIdTypes.mat --> ",str(global.dictIdTypes.mat,4,0)," material\n")
        #debug concat("global.dictIdTypes.den --> ",str(global.dictIdTypes.den,4,0)," density\n")
        #debug concat("global.dictIdTypes.dem --> ",str(global.dictIdTypes.dem,4,0)," density_map\n")
        #debug concat("global.dictIdTypes.med --> ",str(global.dictIdTypes.med,4,0)," media\n")
        #debug concat("global.dictIdTypes.fog --> ",str(global.dictIdTypes.fog,4,0)," fog\n")
        #debug concat("global.dictIdTypes.clm --> ",str(global.dictIdTypes.clm,4,0)," color_map\n")
        #debug concat("global.dictIdTypes.pgm --> ",str(global.dictIdTypes.mat,4,0)," pigment_map\n")
        #debug concat("global.dictIdTypes.trm --> ",str(global.dictIdTypes.trm,4,0)," transform\n")
        #debug concat("global.dictIdTypes.cam --> ",str(global.dictIdTypes.cam,4,0)," camera\n")
        #debug concat("global.dictIdTypes.mcr --> ",str(global.dictIdTypes.mcr,4,0)," macro\n")
        #debug concat("global.dictIdTypes.spl --> ",str(global.dictIdTypes.spl,4,0)," spline\n")
        #debug concat("global.dictIdTypes.sky --> ",str(global.dictIdTypes.sky,4,0)," sky_sphere\n")
        #debug "-------\n"
    #end
#end


//------------------------------------------------------------------------------

#ifndef(Setidtypes_Inc_Temp)
#declare Setidtypes_Inc_Temp = version;
#version 3.8; // For povr adding 'unoffical' makes some sense, but warnings then ugly.

#ifdef(View_POV_Include_Stack)
    #debug "including setidtypes.inc\n"
#end


//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="setidtypes.inc")

    #fopen FID "SelfTestSetIdTypesInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n")
    #write (FID,"\n---\n")
    #write (FID,concat("Macro_Parsing 0.123456789 ",str(9.876543210,19,16)," Assuming >0 of macro calls fail.\n"))

    #debug "===\n"
    McrPrintGlobalDictIdTypes()
    McrPrintGlobalDictIdTypes(true)
    #debug "===\n"
    McrPrintGlobalDictIdTypes(false)
    #debug "===\n"
    McrPrintGlobalDictIdTypes(true)
    #debug "===\n"

    #write (FID,"\n")
    #fclose FID

    #fopen FID "SelfTestSetIdTypesInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n")
    #write (FID,"\n---\n")
    #write (FID,concat("McrPrintGlobalDictIdTypes 1.000000 ",str(1.0,9,6)," null arg\n"))
    #write (FID,concat("McrPrintGlobalDictIdTypes 1.000000 ",str(1.0,9,6)," false arg\n"))
    #write (FID,concat("McrPrintGlobalDictIdTypes 1.000000 ",str(1.0,9,6)," true arg\n"))
    #write (FID,"\n---\n")
    // Run generic object test.
    #declare Thing00 = union{object{sphere{0,1}} object{sphere{1,1}}}
    #if ((id_type(Thing00)=global.dictIdTypes.obj) | (id_type(Thing00)<0.0))
        #write (FID,concat("GenericObject 1.0 1.0 Thing00 is one of POV-Ray's generic object types.\n"))
    #else
        #write (FID,concat("GenericObject 1.0 0.0 Thing00 is one of POV-Ray's generic object types.\n"))
    #end

    #write (FID,"\n")
    #fclose FID

    #error "\n\n===Self testing. Stop before render.===\n"

#end // End self testing

#version Setidtypes_Inc_Temp;
#end //setidtypes.inc


