// This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License.
// To view a copy of this license, visit http://creativecommons.org/licenses/by-sa/3.0/ or send a
// letter to Creative Commons, 444 Castro Street, Suite 900, Mountain View, California, 94041, USA.

// Persistence of Vision Ray Tracer version 3.8 Include File. (povr version ONLY)
// File: vectoranalysis.inc
// Last updated: 28 May 2021

// Description: This file contains various vector analysis macros.

// Modified:
// 16 May 2021  WFP - Eliminated __Gradient_Fn_Accuracy_. Shorter _<lc> povr args.

//------------------------------------------------------------------------------

#ifndef(VECTORANALYSIS_INC_TEMP)
#declare VECTORANALYSIS_INC_TEMP = version;
#version 3.8;   // For povr adding 'unoffical' makes some sense, but warnings then ugly.

#ifdef(View_POV_Include_Stack)
    #debug "including vectoranalysis.inc\n"
#end

#include "functions.inc"  // f_r

// --------------------------------------------------------
// Vector analysis macros
// --------------------------------------------------------
// Authors: Christoph Hormann and Tor Olav Kristensen

// Various functions of vector analysis in form of macros that can be used in
// user defined functions or expressions

// Because vector functions can only be created as pigment or transform/spline
// functions and can not be passed as a macro parameter there is no fn_Curl()
// function and the divergence and curl macros use 3 float functions for
// defining the vector field.


// macro calculating the gradient of a function
// as a function
//
// Parameters:
//     _fn : function to calculate the gradient from
//   Optional
//     _accuracy
//
// Output: the length of the gradient as a function
#macro fn_Gradient(_fn,optional _accuracy)
    #ifndef(local._accuracy)
        #local Accuracy = 0.001;
    #else
        #local Accuracy = _accuracy;
    #end
    function {
        f_r(
            _fn(x + Accuracy, y, z) - _fn(x - Accuracy, y, z),
            _fn(x, y + Accuracy, z) - _fn(x, y - Accuracy, z),
            _fn(x, y, z + Accuracy) - _fn(x, y, z - Accuracy)
        )/(2*Accuracy)
    }
#end


// macro calculating the gradient of a function
// in one direction as a function
//
// Parameters:
//     _fn  : function to calculate the gradient from
//     _dir : direction to calculate the gradient
//   Optional
//     _accuracy
//
// Output: the gradient in that direction as a function
#macro fn_Gradient_Directional(_fn,_dir,optional _accuracy)
    #ifndef(local._accuracy)
        #local Accuracy = 0.001;
    #else
        #local Accuracy = _accuracy;
    #end
    #local Dirx = vnormalize(_dir).x;
    #local Diry = vnormalize(_dir).y;
    #local Dirz = vnormalize(_dir).z;
    function {
        (
            (_fn(x + Accuracy, y, z) - _fn(x - Accuracy, y, z))*Dirx +
            (_fn(x, y + Accuracy, z) - _fn(x, y - Accuracy, z))*Diry +
            (_fn(x, y, z + Accuracy) - _fn(x, y, z - Accuracy))*Dirz
        )/(2*Accuracy)
    }
#end


// macro calculating the divergence of a (vector) function
// as a function
//
// Parameters:
//     _fnx,
//     _fny,
//     _fnz: x, y and z components of a vector function
//   Optional
//     _accuracy
//
// Output: the divergence as a function
#macro fn_Divergence(_fnx,_fny,_fnz,optional _accuracy)
    #ifndef(local._accuracy)
        #local Accuracy = 0.001;
    #else
        #local Accuracy = _accuracy;
    #end
    function {
        (
        _fnx(x + Accuracy, y, z) - _fnx(x - Accuracy, y, z) +
        _fny(x, y + Accuracy, z) - _fny(x, y - Accuracy, z) +
        _fnz(x, y, z + Accuracy) - _fnz(x, y, z - Accuracy)
        )/(2*Accuracy)
    }
#end


// macro calculating the gradient of a function
// as a vector expression
//
// Parameters:
//     _fn : function to calculate the gradient from
//     _p0 : point where to calculate the gradient
//   Optional
//     _accuracy
//
// Output: the gradient as a vector expression
#macro vGradient(_fn,_p0,optional _accuracy)
    #ifndef(local._accuracy)
        #local Accuracy = 0.001;
    #else
        #local Accuracy = _accuracy;
    #end
    #local P0x=_p0.x;
    #local P0y=_p0.y;
    #local P0z=_p0.z;
    (
    <
        _fn(P0x + Accuracy, P0y, P0z) - _fn(P0x - Accuracy, P0y, P0z),
        _fn(P0x, P0y + Accuracy, P0z) - _fn(P0x, P0y - Accuracy, P0z),
        _fn(P0x, P0y, P0z + Accuracy) - _fn(P0x, P0y, P0z - Accuracy)
    >/(2*Accuracy)
    )
#end


// macro calculating the curl of a (vector) function
// as a vector expression
//
// Parameters:
//     _fnx,
//     _fny,
//     _fnz : x, y and z components of a vector function
//     _p0  : point where to calculate the curl
//   Optional
//     _accuracy
//
// Output: the curl as a vector expression
#macro vCurl(_fnx,_fny,_fnz,_p0,optional _accuracy)
    #ifndef(local._accuracy)
        #local Accuracy = 0.001;
    #else
        #local Accuracy = _accuracy;
    #end
    #local P0x=_p0.x;
    #local P0y=_p0.y;
    #local P0z=_p0.z;
    (
    <
        _fnz(P0x, P0y + Accuracy, P0z) - _fnz(P0x, P0y - Accuracy, P0z) -
        _fny(P0x, P0y, P0z + Accuracy) + _fny(P0x, P0y, P0z - Accuracy),

        _fnx(P0x, P0y, P0z + Accuracy) - _fnx(P0x, P0y, P0z - Accuracy) -
        _fnz(P0x + Accuracy, P0y, P0z) + _fnz(P0x - Accuracy, P0y, P0z),

        _fny(P0x + Accuracy, P0y, P0z) - _fny(P0x - Accuracy, P0y, P0z) -
        _fnx(P0x, P0y + Accuracy, P0z) + _fnx(P0x, P0y - Accuracy, P0z)
    >/(2*Accuracy)
    )
#end


// macro calculating the divergence of a (vector) function
// as a float expression
//
// Parameters:
//     _fnx,
//     _fny,
//     _fnz : x, y and z components of a vector function
//     _p0  : point where to calculate the divergence
//   Optional
//     _accuracy
//
// Output: the divergence as a float expression
#macro Divergence(_fnx,_fny,_fnz,_p0,optional _accuracy)
    #ifndef(local._accuracy)
        #local Accuracy = 0.001;
    #else
        #local Accuracy = _accuracy;
    #end
    #local P0x=_p0.x;
    #local P0y=_p0.y;
    #local P0z=_p0.z;
    (
        (
        _fnx(P0x + Accuracy, P0y, P0z) - _fnx(P0x - Accuracy, P0y, P0z) +
        _fny(P0x, P0y + Accuracy, P0z) - _fny(P0x, P0y - Accuracy, P0z) +
        _fnz(P0x, P0y, P0z + Accuracy) - _fnz(P0x, P0y, P0z - Accuracy)
        )/(2*Accuracy)
    )
#end


// macro calculating the length of the gradient
// of a function as a float expression
//
// Parameters:
//     _fn : function to calculate the gradient from
//     _p0 : point where to calculate the gradient
//   Optional
//     _accuracy
//
// Output: the length of the gradient as a float expression
#macro Gradient_Length(_fn,_p0,optional _accuracy)
    #ifndef(local._accuracy)
        #local Accuracy = 0.001;
    #else
        #local Accuracy = _accuracy;
    #end
    vlength(vGradient( function { _fn(x,y,z) },_p0,Accuracy))
#end


// macro calculating the gradient of a function
// in one direction as a float expression
//
// Parameters:
//     _fn  : function to calculate the gradient from
//     _p0  : point where to calculate the gradient
//     _dir : direction to calculate the gradient
//   Optional
//     _accuracy
//
// Output: the gradient in that direction as a float expression
#macro Gradient_Directional(_fn,_p0,_dir,optional _accuracy)
    #ifndef(local._accuracy)
        #local Accuracy = 0.001;
    #else
        #local Accuracy = _accuracy;
    #end
    vdot(
        vGradient( function { _fn(x,y,z) },_p0,Accuracy),
        vnormalize(_dir)
    )
#end

//================================= End render time declarations ===================================

//--- Self testing:  povr +mv3.8 +i<this_file>
#if(input_file_name="vectoranalysis.inc")

 // --- Currently not 'just' calling macros herein.
 // #fopen FID "SelfTestVectorAnalysisInc.txt" write
 // #write (FID,"\nExpect value followed by value generated.\n")
 // // Run some macros to be sure parse works. Assume failure and correct later if all parse OK.
 // #write (FID,"\n---\n")
 // #write (FID,concat("Macro_Parsing 0.123456789 ",str(9.876543210,19,16)," Assuming >0 of macro calls fail.\n"))
 // #fclose FID

    //-------- Now tests for more than just parsing. Will reach if all above parse clean.
    #fopen FID "SelfTestVectorAnalysisInc.txt" write
    #write (FID,"\nExpect value followed by value generated.\n")
 // // If all macro call tests above work create output for what was tested.
 // #write (FID,concat("Macro_Parsing 1.000000 ",str(1.0,9,6)," Center_Object OK\n"))
    #write (FID,"\n")

    // fn_Gradient
    #declare FnSph00 = function { f_sphere(x,y,z,1) }
    #declare Fn00    = fn_Gradient(FnSph00,)
    #write (FID,concat("fn_Gradient 1.000000 ",str(Fn00(0.5,0.5,0.5),9,6)," \n"))
    #declare FnSph01 = function { pow(f_sphere(x,y,z,1),2) }
    #declare Fn01    = fn_Gradient(FnSph01,)
    #write (FID,concat("fn_Gradient 0.267948 ",str(Fn01(0.5,0.5,0.5),9,6)," \n"))
    #ifdef (Fn00) #undef Fn00 #end
    #ifdef (Fn01) #undef Fn01 #end
    #write (FID,"\n")
    // fn_Gradient_Directional
    #declare Fn00    = fn_Gradient_Directional(FnSph00,<1,0,0>,)
    #write (FID,concat("fn_Gradient_Directional +0.577350 ",str(Fn00(0.5,0.5,0.5),9,6)," \n"))
    #declare Fn01    = fn_Gradient_Directional(FnSph01,<1,0,0>,0.001)
    #write (FID,concat("fn_Gradient_Directional -0.154700 ",str(Fn01(0.5,0.5,0.5),9,6)," \n"))
    #ifdef (FnSph00) #undef FnSph00 #end
    #ifdef (FnSph01) #undef FnSph01 #end
    #ifdef (Fn00) #undef Fn00 #end
    #ifdef (Fn01) #undef Fn01 #end
    #write (FID,"\n")
    // fn_Divergence where we have a vector component functions.
    #declare Fnx = function { f_turbulence(x,y,z,0,2,1,0,1,6,0.5,2.0) }
    #declare Fny = function { f_turbulence(x,y,z,1,2,1,0,1,6,0.5,2.0) }
    #declare Fnz = function { f_turbulence(x,y,z,2,2,1,0,1,6,0.5,2.0) }
    #declare Fn00 = fn_Divergence(Fnx,Fny,Fnz,0.002)
    #declare Val = Fn00(1.1,2.2,3.3);
    #write (FID,concat("fn_Divergence +4.041933 ",str(Val,9,6)," \n"))
    #ifdef (Fnx) #undef Fnx #end
    #ifdef (Fny) #undef Fny #end
    #ifdef (Fnz) #undef Fnz #end
    #ifdef (Fn00) #undef Fn00 #end
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // vGradient
    #declare FnSph00 = function { f_sphere(x,y,z,1) }
    #declare Vec00   = vGradient(FnSph00,<0.5,0.5,0.5>,0.001);
    #write (FID,concat("vGradient 0.577350 ",str(Vec00.x,9,6)," Vec00.x\n"))
    #write (FID,concat("vGradient 0.577350 ",str(Vec00.y,9,6)," Vec00.y\n"))
    #write (FID,concat("vGradient 0.577350 ",str(Vec00.z,9,6)," Vec00.z\n"))
    #write (FID,"-\n")
    #declare FnSph01 = function { pow(f_sphere(x,y,z,1),2) }
    #declare Vec01   = vGradient(FnSph01,<0.5,0.5,0.5>,0.001);
    #write (FID,concat("vGradient -0.154700 ",str(Vec01.x,9,6)," Vec01.x\n"))
    #write (FID,concat("vGradient -0.154700 ",str(Vec01.y,9,6)," Vec01.y\n"))
    #write (FID,concat("vGradient -0.154700 ",str(Vec01.z,9,6)," Vec01.z\n"))
    #ifdef (FnSph00) #undef FnSph00 #end
    #ifdef (FnSph01) #undef FnSph01 #end
    #ifdef (Vec00) #undef Vec00 #end
    #ifdef (Vec01) #undef Vec01 #end
    #write (FID,"\n")
    // vCurl
    #declare Fnx = function { f_turbulence(x,y,z,0,2,1,0,1,6,0.5,2.0) }
    #declare Fny = function { f_turbulence(x,y,z,1,2,1,0,1,6,0.5,2.0) }
    #declare Fnz = function { f_turbulence(x,y,z,2,2,1,0,1,6,0.5,2.0) }
    #declare Vec00 = vCurl(Fnx,Fny,Fnz,<1.1,2.2,3.3>,0.002);
    #write (FID,concat("vCurl -0.322579 ",str(Vec00.x,9,6)," Vec00.x \n"))
    #write (FID,concat("vCurl -2.132033 ",str(Vec00.y,9,6)," Vec00.y \n"))
    #write (FID,concat("vCurl -0.077065 ",str(Vec00.z,9,6)," Vec00.z \n"))
    #ifdef (Fnx) #undef Fnx #end
    #ifdef (Fny) #undef Fny #end
    #ifdef (Fnz) #undef Fnz #end
    #ifdef (Vec00) #undef Vec00 #end
    #write (FID,"\n")
    // Divergence
    #declare Fnx = function { f_turbulence(x,y,z,0,2,1,0,1,6,0.5,2.0) }
    #declare Fny = function { f_turbulence(x,y,z,1,2,1,0,1,6,0.5,2.0) }
    #declare Fnz = function { f_turbulence(x,y,z,2,2,1,0,1,6,0.5,2.0) }
    #declare Val = Divergence(Fnx,Fny,Fnz,<1.1,2.2,3.3>,0.002);
    #write (FID,concat("Divergence +4.041933 ",str(Val.x,9,6)," \n"))
    #ifdef (Fnx) #undef Fnx #end
    #ifdef (Fny) #undef Fny #end
    #ifdef (Fnz) #undef Fnz #end
    #ifdef (Val) #undef Val #end
    #write (FID,"\n")
    // Gradient_Length
    #declare FnSph00 = function { f_sphere(x,y,z,1) }
    #declare Var = Gradient_Length(FnSph00,<0.5,0.5,0.5>,);
    #write (FID,concat("Gradient_Length +1.000000 ",str(Var,9,6)," \n"))
    #declare FnSph01 = function { pow(f_sphere(x,y,z,1),2) }
    #declare Var = Gradient_Length(FnSph01,<0.5,0.5,0.5>,0.001);
    #write (FID,concat("Gradient_Length +0.267948 ",str(Var,9,6)," \n"))
    #ifdef (FnSph00) #undef FnSph00 #end
    #ifdef (FnSph01) #undef FnSph01 #end
    #ifdef (Var) #undef Var #end
    #write (FID,"\n")
    // Gradient_Directional
    #declare FnSph00 = function { f_sphere(x,y,z,1) }
    #declare Var = Gradient_Directional(FnSph00,<0.5,0.5,0.5>,x,);
    #write (FID,concat("Gradient_Directional +0.577350 ",str(Var,9,6)," \n"))
    #declare FnSph01 = function { pow(f_sphere(x,y,z,1),2) }
    #declare Var = Gradient_Directional(FnSph01,<0.7,0.7,0.7>,x,0.001);
    #write (FID,concat("Gradient_Directional +0.245300 ",str(Var,9,6)," \n"))
    #ifdef (FnSph00) #undef FnSph00 #end
    #ifdef (FnSph01) #undef FnSph01 #end
    #ifdef (Var) #undef Var #end
    #write (FID,"\n")

    #fclose FID

    #error "\n\n===Self testing. Stop before render.===\n"

#end // End self testing

#version VECTORANALYSIS_INC_TEMP;
#end //vectoranalysis.inc

