//******************************************************************************
///
/// @file core/shape/parametric.cpp
///
/// Implementation of the parametric geometric primitive.
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2021 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//------------------------------------------------------------------------------
// SPDX-License-Identifier: AGPL-3.0-or-later
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "core/shape/parametric.h"

// C++ variants of C standard header files
//  (none at the moment)

// C++ standard header files
#include <algorithm>
#include <array>

// POV-Ray header files (base module)
//  (none at the moment)

// POV-Ray header files (core module)
#include "core/math/matrix.h"
#include "core/render/ray.h"
#include "core/scene/tracethreaddata.h"
#include "core/support/statistics.h"

// this must be the last file included
#include "base/povdebug.h"

namespace pov
{

using std::min;
using std::max;

// Local preprocessor defines

const int Max_intNumber = 10000000;
#define close(x, y) (fabs(x-y) < gkMinIsectDepthReturned ? 1 : 0)

const int INDEX_U = 0;
const int INDEX_V = 1;

const int OK_X     =   1;
const int OK_Y     =   2;
const int OK_Z     =   4;
const int OK_R     =   8;
const int OK_S     =  16;
const int OK_T     =  32;
const int OK_U     =  64;
const int OK_V     = 128;


bool Parametric::All_Intersections(const Ray& ray, IStack& Depth_Stack, TraceThreadData *Thread)
{
    Vector3d P, D, IPoint;
    Vector2d low_vect, hi_vect, uv;
    BasicRay New_Ray;
    double XRayMin, XRayMax, YRayMin, YRayMax, ZRayMin, ZRayMax;
    double TPotRes, TLen;  // Read as ThisPotentialResult, ThisLength and below ThisResult.
    double Depth1, Depth2, temp, Len, TResult = HUGE_VAL;
    double lo, hi;
    int MaxPrecompX, MaxPrecompY, MaxPrecompZ;
    int split, i = 0, Side1, Side2;
    int parX, parY;
    double Intervals_Low[2][32];
    double Intervals_Hi[2][32];
    int SectorNum[32];

    Thread->Stats()[Ray_Par_Bound_Tests]++;

    if (!container->Intersect(ray, Trans, Depth1, Depth2, Side1, Side2))
        return false;

    Thread->Stats()[Ray_Par_Bound_Tests_Succeeded]++;
    Thread->Stats()[Ray_Parametric_Tests]++;

    if (Trans != nullptr)
    {
        MInvTransPoint(P, ray.Origin, Trans);
        MInvTransDirection(D, ray.Direction, Trans);
    }
    else
    {
        P = ray.Origin;
        D = ray.Direction;
    }

    if (Depth1 == Depth2)
        Depth1 = 0;

    if ((Depth1 += 4 * accuracy) > Depth2)
        return false;

    GenericScalarFunctionInstance fn0(Function[0], Thread);
    GenericScalarFunctionInstance fn1(Function[1], Thread);
    GenericScalarFunctionInstance fn2(Function[2], Thread);

    Intervals_Low[INDEX_U][0] = umin;
    Intervals_Hi[INDEX_U][0] = umax;

    Intervals_Low[INDEX_V][0] = vmin;
    Intervals_Hi[INDEX_V][0] = vmax;
    /* Fri 09-27-1996 0. */
    SectorNum[0] = 1;

    MaxPrecompX = MaxPrecompY = MaxPrecompZ = 0;
    if (PData != nullptr)
    {
        if (((PData->flags) & OK_X) != 0)
            MaxPrecompX = 1 << (PData->depth);
        if (((PData->flags) & OK_Y) != 0)
            MaxPrecompY = 1 << (PData->depth);
        if (((PData->flags) & OK_Z) != 0)
            MaxPrecompZ = 1 << (PData->depth);
    }

    /* 0 */
    while (i >= 0)
    {
        low_vect[U] = Intervals_Low[INDEX_U][i];
        hi_vect[U] = Intervals_Hi[INDEX_U][i];
        Len = hi_vect[U] - low_vect[U];
        split = INDEX_U;

        low_vect[V] = Intervals_Low[INDEX_V][i];
        hi_vect[V] = Intervals_Hi[INDEX_V][i];
        temp = hi_vect[V] - low_vect[V];
        if (temp > Len)
        {
            Len = temp;
            split = INDEX_V;
        }
        parX = parY = 0;
        TLen = 0;

        /* X */
        if (SectorNum[i] < MaxPrecompX)
        {
            lo = PData->Low[0][SectorNum[i]];
            hi = PData->Hi[0][SectorNum[i]];
        }
        else
            Evaluate_Function_Interval_UV(fn0, accuracy, low_vect, hi_vect, max_gradient, lo, hi);
        /* fabs(D.x() *(T2-T1)) is not OK with new method */

        if (close(D.x(),0))                    // No meaningful change in X direction.
        {
            parX = 1;                          // Let the Y and Z code know ray traveling parallel to PX.
            if ((hi < P.x()) || (lo > P.x()))  // If either true, ray misses surface
            {
                i--;
                continue;
            }
        }
        else
        {
            XRayMin = (hi - P.x()) / D.x();    // Normalized distance from ray x to current hi
            XRayMax = (lo - P.x()) / D.x();    // Normalized distance from ray x to current hi
            if (XRayMin > XRayMax)             // If out of order(?), swap.
            {
                temp = XRayMin;
                XRayMin = XRayMax;
                XRayMax = temp;
            }

            if ((XRayMin > Depth2) || (XRayMax < Depth1))  // Range outside initial bounding.
            {
                i--;
                continue;
            }

            if ((TPotRes = XRayMin) > TResult) // Performance gain of this is significant.
            {
                i--;
                continue;
            }

            TLen = XRayMax - XRayMin;         // Calculate ThisLength.
        }

        /* Y */
        if (SectorNum[i] < MaxPrecompY)
        {
            lo = PData->Low[1][SectorNum[i]];
            hi = PData->Hi[1][SectorNum[i]];
        }
        else
            Evaluate_Function_Interval_UV(fn1, accuracy, low_vect, hi_vect, max_gradient, lo, hi);

        if (close(D.y(), 0))                   // No meaningful change in X direction.
        {
            parY = 1;                          // Let the Z code know ray traveling parallel to PY.
            if ((hi < P.y()) || (lo > P.y()))  // If either true, ray misses surface.
            {
                i--;
                continue;
            }
        }
        else
        {
            YRayMin = (hi - P.y()) / D.y();    // Normalized distance from ray y to current hi
            YRayMax = (lo - P.y()) / D.y();    // Normalized distance from ray y to current lo
            if (YRayMin > YRayMax)             // If out of order(?), swap.
            {
                temp = YRayMin;
                YRayMin = YRayMax;
                YRayMax = temp;
            }

            if ((YRayMin > Depth2) || (YRayMax < Depth1))  // Range outside initial bounding.
            {
                i--;
                continue;
            }

            if ((TPotRes = YRayMin) > TResult) // Performance gain of this is significant.
            {
                i--;
                continue;
            }

            if (parX == 0)  // If ray not parallel to PX, range short or past the 'first' surface. (Bug?)
            {
                if ((YRayMin > XRayMax) || (YRayMax < XRayMin))
                {
                    i--;
                    continue;
                }
            }

            if ((temp = YRayMax - YRayMin) > TLen)  // If this Y length greater than current ThisLength, use it (larger).
                TLen = temp;
        }

        /* Z */
        if ((SectorNum[i] < MaxPrecompZ) && (0 < SectorNum[i]))
        {
            lo = PData->Low[2][SectorNum[i]];
            hi = PData->Hi[2][SectorNum[i]];
        }
        else
            Evaluate_Function_Interval_UV(fn2, accuracy, low_vect, hi_vect, max_gradient, lo, hi);

        if (close(D.z(), 0))                   // No meaningfule change in Z direction.
        {
            if ((hi < P.z()) || (lo > P.z()))  // If either true, ray misses surface.
            {
                i--;
                continue;
            }
        }
        else
        {
            ZRayMin = (hi - P.z()) / D.z();    // Normalized distance from ray z to current hi
            ZRayMax = (lo - P.z()) / D.z();    // Normalized distance from ray z to current lo
            if (ZRayMin > ZRayMax)             // If out of order(?), swap?
            {
                temp = ZRayMin;
                ZRayMin = ZRayMax;
                ZRayMax = temp;
            }

            if ((ZRayMin > Depth2) || (ZRayMax < Depth1))  // Range outside initial bounding.
            {
                i--;
                continue;
            }

            if ((TPotRes = ZRayMin) > TResult) // Performance gain of this is significant.
            {
                i--;
                continue;
            }

            if (parX == 0)  // If ray not parallel to PX, range short or past the 'first' surface. (Bug?)
            {
                if ((ZRayMin > XRayMax) || (ZRayMax < XRayMin))
                {
                    i--;
                    continue;
                }
            }

            if (parY == 0)  // If ray not parallel to PY, range short or past the 'first' surface. (Bug?)
            {
                if ((ZRayMin > YRayMax) || (ZRayMax < YRayMin))
                {
                    i--;
                    continue;
                }
            }

            if ((temp = ZRayMax - ZRayMin) > TLen)  // If this Z length greater than current ThisLength, use it (larger).
                TLen = temp;
        }

        if (Len > TLen)
            Len = TLen;  // If ThisLength is smaller than current shortest Length, update Length.

        if (Len < accuracy) // All x,y,z value range lengths are less than user's desired accuracy. Good to go.
        {
            if ((TResult > TPotRes) && (TPotRes > Depth1))  // If ThisPotentialResult is beyond low bounding and better
            {                                               // than ThisResult, update ThisResult. Bail to better things.
                TResult = TPotRes;
                uv = low_vect;
            }
            i--;
        }
        else
        {
            /* 1 copy */
            if ((SectorNum[i] *= 2) >= Max_intNumber)
                SectorNum[i] = Max_intNumber;
            SectorNum[i + 1] = SectorNum[i];
            SectorNum[i]++;
            i++;
            Intervals_Low[INDEX_U][i] = low_vect[U];
            Intervals_Hi[INDEX_U][i] = hi_vect[U];

            Intervals_Low[INDEX_V][i] = low_vect[V];
            Intervals_Hi[INDEX_V][i] = hi_vect[V];

            /* 2 split */
            temp = (Intervals_Hi[split][i] + Intervals_Low[split][i]) / 2.0;
            Intervals_Hi[split][i] = temp;
            Intervals_Low[split][i - 1] = temp;
        }
    }

    if ((TResult < Depth2) && (TResult > 0.0))
    {
        Thread->Stats()[Ray_Parametric_Tests_Succeeded]++;
        IPoint = ray.Direction * TResult;
        IPoint += ray.Origin;

        if (Clip.empty() || Point_In_Clip(IPoint, Clip, Thread))
        {
            // compute_param_normal( Par, UResult, VResult , &N);
            // push_normal_entry( TResult ,IPoint, N, reinterpret_cast<ObjectPtr>(Object), Depth_Stack);
            Depth_Stack->push(Intersection(TResult, IPoint, uv, this));

            return true;
        }
    }

    return false;
}
/* 0 */


bool Parametric::Inside(const Vector3d&, TraceThreadData *Thread) const
{
    return false;
}


void Parametric::Normal(Vector3d& Result, Intersection *Inter, TraceThreadData *Thread) const
{
    Vector3d RU, RV;
    Vector2d uv_vect;

    std::array<GenericScalarFunctionInstance,3> aFn = {
        GenericScalarFunctionInstance(Function[0], Thread),
        GenericScalarFunctionInstance(Function[1], Thread),
        GenericScalarFunctionInstance(Function[2], Thread)
    };

    uv_vect[U] = Inter->Iuv[U];
    uv_vect[V] = Inter->Iuv[V];
    RU[X] = RV[X] = -aFn[X].Evaluate(uv_vect);
    RU[Y] = RV[Y] = -aFn[Y].Evaluate(uv_vect);
    RU[Z] = RV[Z] = -aFn[Z].Evaluate(uv_vect);

    uv_vect[U] += accuracy;
    RU[X] += aFn[X].Evaluate(uv_vect);
    RU[Y] += aFn[Y].Evaluate(uv_vect);
    RU[Z] += aFn[Z].Evaluate(uv_vect);

    uv_vect[U] = Inter->Iuv[U];
    uv_vect[V] += accuracy;
    RV[X] += aFn[X].Evaluate(uv_vect);
    RV[Y] += aFn[Y].Evaluate(uv_vect);
    RV[Z] += aFn[Z].Evaluate(uv_vect);

    Result = cross(RU, RV);
    if (Trans != nullptr)
        MTransNormal(Result, Result, Trans);
    Result.normalize();
}


void Parametric::Compute_BBox()
{
    container->ComputeBBox(BBox);
    if (Trans != nullptr)
    {
        Recompute_BBox(&BBox, Trans);
    }
}


void Parametric::Translate(const Vector3d&, const TRANSFORM* tr)
{
    Transform(tr);
}


void Parametric::Rotate(const Vector3d&, const TRANSFORM* tr)
{
    Transform(tr);
}


void Parametric::Scale(const Vector3d&, const TRANSFORM* tr)
{
    Transform(tr);
}


void Parametric::Transform(const TRANSFORM* tr)
{
    if (Trans == nullptr)
        Trans = Create_Transform();
    Compose_Transforms(Trans, tr);
    Compute_BBox();
}


ObjectPtr Parametric::Copy()
{
    Parametric *New = new Parametric();
    Destroy_Transform(New->Trans);
    *New = *this;

    New->Function[0] = Function[0]->Clone();
    New->Function[1] = Function[1]->Clone();
    New->Function[2] = Function[2]->Clone();
    New->Trans = Copy_Transform(Trans);
    New->PData = Copy_PrecompParVal();

    New->container = std::shared_ptr<ContainedByShape>(container->Copy());

    return (New);
}


Parametric::~Parametric()
{
    delete Function[0];
    delete Function[1];
    delete Function[2];
    Destroy_PrecompParVal();
}


Parametric::Parametric() : NonsolidObject(PARAMETRIC_OBJECT)
{
    container = std::shared_ptr<ContainedByShape>(new ContainedByBox());

    Make_BBox(BBox, -1.0, -1.0, -1.0, 2.0, 2.0, 2.0);

    Trans = Create_Transform();

    Function[0] = nullptr;
    Function[1] = nullptr;
    Function[2] = nullptr;
    accuracy = 0.001;
    max_gradient = 1;
    PData = nullptr;
}


void Parametric::UVCoord(Vector2d& Result, const Intersection *inter) const
{
    Result = inter->Iuv;
}


void Parametric::Precomp_Par_Int(int depth, double umin, double vmin, double umax, double vmax,
        GenericScalarFunctionInstance aFn[3])
{
    int j;

    if (depth >= PrecompLastDepth)
    {
        for (j = 0; j < 3; j++)
        {
            if (PData->flags & (1 << j))
            {
                Vector2d low,hi;

                low[U] = umin;
                hi[U]  = umax;
                low[V] = vmin;
                hi[V]  = vmax;

                Evaluate_Function_Interval_UV(aFn[j],
                                              accuracy,
                                              low,
                                              hi,
                                              max_gradient,
                                              PData->Low[j][depth],
                                              PData->Hi[j][depth]);
            }
        }
    }
    else                    /* split */
    {
        if (umax - umin < vmax - vmin)
        {
            Precomp_Par_Int(2 * depth, umin, vmin, umax, (vmin + vmax) / 2.0, aFn);
            Precomp_Par_Int(2 * depth + 1, umin, (vmin + vmax) / 2.0, umax, vmax, aFn);
        }
        else
        {
            Precomp_Par_Int(2 * depth, umin, vmin, (umin + umax) / 2.0, vmax, aFn);
            Precomp_Par_Int(2 * depth + 1, (umin + umax) / 2.0, vmin, umax, vmax, aFn);
        }
        for (j = 0; j < 3; j++)
        {
            if (PData->flags & (1 << j))
            {
                if (PData->Hi[j][2 * depth] > PData->Hi[j][2 * depth + 1])
                    PData->Hi[j][depth] = PData->Hi[j][2 * depth];
                else
                    PData->Hi[j][depth] = PData->Hi[j][2 * depth + 1];
                if (PData->Low[j][2 * depth] < PData->Low[j][2 * depth + 1])
                    PData->Low[j][depth] = PData->Low[j][2 * depth];
                else
                    PData->Low[j][depth] = PData->Low[j][2 * depth + 1];
            }
        }
    }
}


void Parametric::Precompute_Parametric_Values(char flags, int depth, TraceThreadData *Thread)
{
    const char* es = "precompute";
    int nmb;

    if ((depth < 1) || (depth > 20))
        throw POV_EXCEPTION_STRING("Precompute: invalid depth");
    nmb = 1 << depth;

    PData = new PRECOMP_PAR_DATA;
    PData->flags = flags;
    PData->depth = depth;
    PData->use = 1;

    if (flags & OK_X)
    {
        PData->Low[0] = new double[nmb];
        PData->Hi[0]  = new double[nmb];
    }
    if (flags & OK_Y)
    {
        PData->Low[1] = new double[nmb];
        PData->Hi[1]  = new double[nmb];
    }
    if (flags & OK_Z)
    {
        PData->Low[2] = new double[nmb];
        PData->Hi[2]  = new double[nmb];
    }

    PrecompLastDepth = 1 << (depth - 1);
    std::array<GenericScalarFunctionInstance,3> aFn = {
        GenericScalarFunctionInstance(Function[0], Thread),
        GenericScalarFunctionInstance(Function[1], Thread),
        GenericScalarFunctionInstance(Function[2], Thread)
    };
    Precomp_Par_Int(1, umin, vmin, umax, vmax, aFn.data());
}


PRECOMP_PAR_DATA* Parametric::Copy_PrecompParVal()
{
    if (PData == nullptr)
        return nullptr;

    (PData->use)++;

    return (PData);
}


void Parametric::Destroy_PrecompParVal()
{
    if (PData == nullptr)
        return;

    PData->use--;
    if (PData->use == 0)
    {
        if (PData->flags & OK_X)
        {
            delete[] PData->Low[0];
            delete[] PData->Hi[0];
        }
        if (PData->flags & OK_Y)
        {
            delete[] PData->Low[1];
            delete[] PData->Hi[1];
        }
        if (PData->flags & OK_Z)
        {
            delete[] PData->Low[2];
            delete[] PData->Hi[2];
        }
        delete PData;
    }
}


// Author Ron Parker
//
// Assume that the function attains its maximum gradient over the whole range
// and determine its minimum and maximum (really lower and upper bounds, not
// strict maxima and minima) accordingly.
//
void Parametric::Interval(double dx, double a, double b, double max_gradient, double *Min, double *Max)
{
    double dy = fabs(a-b);
    double ofs = max_gradient*(dx-dy/max_gradient)/2;
    if ( ofs < 0 ) {
        ofs=0;
    }
    *Max = max(a,b)+ofs;
    *Min = min(a,b)-ofs;
}


void Parametric::Evaluate_Function_Interval_UV(GenericScalarFunctionInstance& fn,
        double threshold, const Vector2d& fnvec_low, const Vector2d& fnvec_hi,
        double max_gradient, double& low, double& hi)
{
    double f_0_0, f_0_1, f_1_0, f_1_1;
    double f_0_min, f_0_max;
    double f_1_min, f_1_max;
    double junk;

    /* Calculate the values at each corner */

    f_0_0 = fn.Evaluate(fnvec_low[U], fnvec_low[V]) - threshold;
    f_0_1 = fn.Evaluate(fnvec_low[U], fnvec_hi[V])  - threshold;
    f_1_0 = fn.Evaluate(fnvec_hi[U],  fnvec_low[V]) - threshold;
    f_1_1 = fn.Evaluate(fnvec_hi[U],  fnvec_hi[V])  - threshold;

    /* Determine a min and a max along the left edge of the patch */
    Interval( fnvec_hi[V]-fnvec_low[V], f_0_0, f_0_1, max_gradient, &f_0_min, &f_0_max);

    /* Determine a min and a max along the right edge of the patch */
    Interval( fnvec_hi[V]-fnvec_low[V], f_1_0, f_1_1, max_gradient, &f_1_min, &f_1_max);

    /* Assume that the upper bounds of both edges are attained at the same
     u coordinate and determine what an upper bound along that line would
     be if it existed.  That's the worst-case maximum value we can reach. */
    Interval( fnvec_hi[U]-fnvec_low[U], f_0_max, f_1_max, max_gradient, &junk, &hi);

    /* same as above to get a lower bound from the two edge lower bounds */
    Interval( fnvec_hi[U]-fnvec_low[U], f_0_min, f_1_min, max_gradient, &low, &junk);

    /*
    char str[200];
    int its=0;
    its++;
    if (its>20) Error("Boom!");
    sprintf( str, "%lf     %lf %lf %lf %lf   %lf %lf\n%lf %lf %lf %lf    %lf %lf    %lf %lf    %lf %lf\n",
        max_gradient,
        fnvec_low[U], fnvec_low[V],
        fnvec_hi[U], fnvec_hi[V],
        fnvec_hi[U]-fnvec_low[U],
        fnvec_hi[V]-fnvec_low[V],

        f_0_0,f_0_1,f_1_0,f_1_1,
        f_0_min,f_0_max,
        f_1_min,f_1_max,
        low,hi);
    Warning( 0, str );
    */
    /* f_min and f_max now contain the maximum interval. */
}

}
// end of namespace pov
